//-------------------------------------------------------------------
//-------------------------------------------------------------------
//
// Simple Spring Mass System
// -- Matrix lib
//
// Primary Author: James F. O'Brien (obrienj@cc.gatech.edu)
//
// (C) Copyright James F. O'Brien, 1995
// (C) Copyright Georgia Institute of Technology, 1995
//-------------------------------------------------------------------
//-------------------------------------------------------------------
//
// RCS Revision History
//
// $Log: bfastMatrix.H,v $
// Revision 1.1  2006/06/16 21:07:14  adamb
// new version
//
// Revision 1.2  2004/03/10 07:41:38  adamb
// fixing warnings in shell simulator
//
// Revision 3.3  2003/10/23 23:35:13  job
// Added trace(.)
//
// Revision 3.2  2003/10/13 20:23:57  job
// Finished port to Linux (Redhat 7.2).
// Moderate amount of testing, things appear to work.
//
// Revision 3.1  2003/10/11 04:08:42  job
// Rough port to Linux
//
// Revision 3.0  1999/03/11 22:29:14  obrienj
// Move from O32 ABI to n32
//
// Revision 2.7  1998/12/20  02:02:27  obrienj
// Eigen decomp and Matrix product added
//
// Revision 2.6  1998/06/10  04:18:10  obrienj
// More restructuring.
// Added SmMatrix2x2 and moved rotation stuff into its own file.
//
// Revision 2.5  1998/02/19  20:56:34  obrienj
// Major restructuring, and addition of some operators.
//
// Revision 2.4  1997/10/25  22:42:19  obrienj
// Added += and -= operator for SmMatrix.
//
// Revision 2.3  1997/10/08  22:50:58  obrienj
// Extended moment and quaternion support.
//
// Revision 2.2  1997/10/03  01:06:02  obrienj
// SmMatrix3x3 inits to identity.
// Added:  SmEulerAngToMatrixXYZ
//         SmMatrixToEulerAngXYZ
//         SmQuaternionToMatrix
//         SmMatrixToQuaternion
// Added:  operator + and - for SmMatrix3x3
//
// Revision 2.1  1997/01/17  08:30:23  obrienj
// Added operator != .
//
// Revision 1.2  1996/09/10  07:23:21  obrienj
// Fixed bug in operaror>> for SmMatrix3x3
// Minor change in format for SmMatrix3x3
// Added mat *= double operator
// Added mat * double operator
//
// Revision 1.1  1996/09/10  01:43:44  obrienj
// Initial revision
//
//
//-------------------------------------------------------------------
//-------------------------------------------------------------------

#ifndef BfastMATRIX_is_defined
#define BfastMATRIX_is_defined


//-------------------------------------------------------------------
//-------------------------------------------------------------------
// Classes from this header:

class BfastMatrix3x3;
class BfastMatrix2x2;

// External classes

class BfastVector3;
class BfastVector2;

#include <iostream>
#include "bfastRotations.H"

//-------------------------------------------------------------------
//-------------------------------------------------------------------

class BfastMatrix3x3 {
  // Stores a 3x3 row ordered matrix. Vectors are assumed to be row
  // vectors and matricies are indexed (row,col).  NOTE: The *
  // operator is matrix multiplication, not a component-wise
  // operation.


public:

  //----------------------------------------------
  // Constructors

  enum ordering { row , col };

  inline BfastMatrix3x3();
  inline BfastMatrix3x3(double x);
  inline BfastMatrix3x3(const BfastMatrix3x3 &that);

  inline BfastMatrix3x3(double xx, double xy, double xz,
		     double yx, double yy, double yz,
		     double zx, double zy, double zz);

  inline BfastMatrix3x3(const double data[3][3], ordering o = row);
  inline BfastMatrix3x3(const BfastVector3 data[3], ordering o = row);
  inline BfastMatrix3x3(const BfastVector3 a      ,
		     const BfastVector3 b      ,
		     const BfastVector3 c      , ordering o = row);

  //----------------------------------------------
  // Index operators

  inline double &operator()(unsigned int r,unsigned int c)      ;
  inline double  operator()(unsigned int r,unsigned int c) const;


  //----------------------------------------------
  // Assignment and set

  inline BfastMatrix3x3 &set      (const double d);
  inline BfastMatrix3x3 &operator=(const double d);

  inline BfastMatrix3x3 &set      (const BfastMatrix3x3 &that);  
  inline BfastMatrix3x3 &operator=(const BfastMatrix3x3 &that); 


  //----------------------------------------------
  // Comparison operators

  inline int operator==(double d) const;
  inline int operator!=(double d) const;
  
  inline int operator==(const BfastMatrix3x3 &that)const; 
  inline int operator!=(const BfastMatrix3x3 &that)const; 

  
  //----------------------------------------------
  // In place arithmetic

  inline BfastMatrix3x3 &operator+=(double d);
  inline BfastMatrix3x3 &operator-=(double d);
  inline BfastMatrix3x3 &operator*=(double d);
  inline BfastMatrix3x3 &operator/=(double d);


  inline BfastMatrix3x3 &operator+=(const BfastMatrix3x3 &that);
  inline BfastMatrix3x3 &operator-=(const BfastMatrix3x3 &that);
  // These are all component-wise operations.

  BfastMatrix3x3 &inplaceMultPre (const BfastMatrix3x3 &that);
  BfastMatrix3x3 &inplaceMultPost(const BfastMatrix3x3 &that);
  // Pre : this = that x this  
  // Post: this = this x that

  inline BfastMatrix3x3 &componentAdd (const BfastMatrix3x3 &that);
  inline BfastMatrix3x3 &componentSub (const BfastMatrix3x3 &that);
  inline BfastMatrix3x3 &componentMult(const BfastMatrix3x3 &that);
  inline BfastMatrix3x3 &componentDiv (const BfastMatrix3x3 &that);

  inline BfastMatrix3x3 &setIdentity     ();
  inline BfastMatrix3x3 &inplaceTranspose();

  //----------------------------------------------
  // Static methods

  inline static unsigned int cycleAxis(unsigned int axis, int direction);


  //----------------------------------------------
  // Define Components

  enum Index { X = 0 , Y = 1 , Z = 2 ,
	       U = 0 , V = 1 , W = 2 };


public:
  //----------------------------------------------
  // Accessable data members

  double data[3][3];

};
  


//-------------------------------------------------------------------
//-------------------------------------------------------------------

class BfastMatrix2x2 {
  // Stores a 2x2 row ordered matrix. Vectors are assumed to be row
  // vectors and matricies are indexed (row,col).  NOTE: The *
  // operator is matrix multiplication, not a component-wise
  // operation.

public:

  //----------------------------------------------
  // Constructors

  enum ordering { row , col };

  inline BfastMatrix2x2();
  inline BfastMatrix2x2(double x);
  inline BfastMatrix2x2(const BfastMatrix2x2 &that);

  inline BfastMatrix2x2(double xx, double xy, 
		     double yx, double yy);

  inline BfastMatrix2x2(const double data[2][2], ordering o = row);
  inline BfastMatrix2x2(const BfastVector2 data[2], ordering o = row);
  inline BfastMatrix2x2(const BfastVector2 a      ,
		     const BfastVector2 b      , ordering o = row);


  //----------------------------------------------
  // Index operators

  inline double &operator()(unsigned int r,unsigned int c)      ;
  inline double  operator()(unsigned int r,unsigned int c) const;


  //----------------------------------------------
  // Assignment and set

  inline BfastMatrix2x2 &set      (const double d);
  inline BfastMatrix2x2 &operator=(const double d);

  inline BfastMatrix2x2 &set      (const BfastMatrix2x2 &that);  
  inline BfastMatrix2x2 &operator=(const BfastMatrix2x2 &that); 


  //----------------------------------------------
  // Comparison operators

  inline int operator!=(const BfastMatrix2x2 &that)const; 
  inline int operator==(const BfastMatrix2x2 &that)const; 

  inline int operator==(double d) const;
  inline int operator!=(double d) const;
  
  
  //----------------------------------------------
  // In place arithmetic

  inline BfastMatrix2x2 &operator+=(double d);
  inline BfastMatrix2x2 &operator-=(double d);
  inline BfastMatrix2x2 &operator*=(double d);
  inline BfastMatrix2x2 &operator/=(double d);


  inline BfastMatrix2x2 &operator+=(const BfastMatrix2x2 &that);
  inline BfastMatrix2x2 &operator-=(const BfastMatrix2x2 &that);
  // These are all component-wise operations.

  BfastMatrix2x2 &inplaceMultPre (const BfastMatrix2x2 &that);
  BfastMatrix2x2 &inplaceMultPost(const BfastMatrix2x2 &that);
  // Pre : this = that x this  
  // Post: this = this x that

  inline BfastMatrix2x2 &componentAdd (const BfastMatrix2x2 &that);
  inline BfastMatrix2x2 &componentSub (const BfastMatrix2x2 &that);
  inline BfastMatrix2x2 &componentDiv (const BfastMatrix2x2 &that);
  inline BfastMatrix2x2 &componentMult(const BfastMatrix2x2 &that);

  inline BfastMatrix2x2 &setIdentity     ();
  inline BfastMatrix2x2 &inplaceTranspose();

  //----------------------------------------------
  // Static methods

  inline static unsigned int cycleAxis(unsigned int axis, int direction);


  //----------------------------------------------
  // Define Components

  enum Index { X = 0 , Y = 1 ,
	       U = 0 , V = 1 };


public:
  //----------------------------------------------
  // Accessable data members

  double data[2][2];

};
  

//-------------------------------------------------------------------
//-------------------------------------------------------------------
//-------------------------------------------------------------------
//-------------------------------------------------------------------
// Operators for class BfastMatrix3x3

inline BfastMatrix3x3 operator+(const BfastMatrix3x3 &a, double b);
inline BfastMatrix3x3 operator-(const BfastMatrix3x3 &a, double b);
inline BfastMatrix3x3 operator*(const BfastMatrix3x3 &a, double b);
inline BfastMatrix3x3 operator/(const BfastMatrix3x3 &a, double b);

inline BfastMatrix3x3 operator+(double a, const BfastMatrix3x3 &b);
inline BfastMatrix3x3 operator-(double a, const BfastMatrix3x3 &b);
inline BfastMatrix3x3 operator*(double a, const BfastMatrix3x3 &b);
inline BfastMatrix3x3 operator/(double a, const BfastMatrix3x3 &b);

inline BfastMatrix3x3 operator+(const BfastMatrix3x3 &a, const BfastMatrix3x3 &b);
inline BfastMatrix3x3 operator-(const BfastMatrix3x3 &a, const BfastMatrix3x3 &b);
// These are component-wise operations. 

inline BfastMatrix3x3 operator*(const BfastMatrix3x3 &a, const BfastMatrix3x3 &b); 
inline BfastVector3   operator*(const BfastMatrix3x3 &a, const BfastVector3   &b);
inline BfastVector3   operator*(const BfastVector3   &a, const BfastMatrix3x3 &b);
// Matrix multiplication.  Vector returning routines treat the vector
// argument like a row or column vector as appropriate.


//-------------------------------------------------------------------
//-------------------------------------------------------------------
// Global funtions for class BfastMatrix3x3

inline double determinant(const BfastMatrix3x3 &a);

inline double trace(const BfastMatrix3x3 &a);


inline BfastMatrix3x3 transpose(const BfastMatrix3x3 &a);
/*  */ BfastMatrix3x3   inverse(const BfastMatrix3x3 &a);

inline BfastMatrix3x3 diagonal(const BfastVector3 &a);


void BfastSVDecomp(const BfastMatrix3x3 &a,
		/* */ BfastMatrix3x3 &u,
		/* */ BfastVector3   &w,
		/* */ BfastMatrix3x3 &v);
// Computes the singular value decomposition of a matrix, [u,w,v].
// The input matrix, a, is related to the output by a =
// u*diagonal(w)*transpose(v).  Both u and v are orthonormal.

void BfastSymetricEigenDecomp(const BfastMatrix3x3 &a,
			   /* */ BfastVector3   &vals,
			   /* */ BfastMatrix3x3 &vecs);
// Given a SYMETRIC matrix, a, it returns the Eigen Values in vals,
// and the Eigen Vectors as the COLUMNS of vecs.  The vectors are
// normalized.  The values/vectors are not sorted in any order.

inline BfastMatrix3x3 BfastMatrixProduct(const BfastVector3 a,const BfastVector3 b);
// Returns the matrix given by a * aT  [ m_ij = a_i * b_j ]

inline BfastMatrix3x3 BfastEigenMatrix(const BfastVector3 a);
// Returns BfastMatrixProduct(a,a)/|a| [ m_ij = a_i * a_j / (a_k * a_k)].
// Note that the (only) eigen vector of the matrix is a, and the
// matrix is symetric.  (Involves a call to sqrt.)


std::istream &operator>>(std::istream &strm,      BfastMatrix3x3 &m);
std::ostream &operator<<(std::ostream &strm,const BfastMatrix3x3 &m);



//-------------------------------------------------------------------
//-------------------------------------------------------------------
//-------------------------------------------------------------------
//-------------------------------------------------------------------
// Operators for class BfastMatrix2x2

inline BfastMatrix2x2 operator+(const BfastMatrix2x2 &a, double b);
inline BfastMatrix2x2 operator-(const BfastMatrix2x2 &a, double b);
inline BfastMatrix2x2 operator*(const BfastMatrix2x2 &a, double b);
inline BfastMatrix2x2 operator/(const BfastMatrix2x2 &a, double b);

inline BfastMatrix3x3 operator+(double a, const BfastMatrix3x3 &b);
inline BfastMatrix3x3 operator-(double a, const BfastMatrix3x3 &b);
inline BfastMatrix3x3 operator*(double a, const BfastMatrix3x3 &b);
inline BfastMatrix3x3 operator/(double a, const BfastMatrix3x3 &b);


inline BfastMatrix2x2 operator+(const BfastMatrix2x2 &a, const BfastMatrix2x2 &b);
inline BfastMatrix2x2 operator-(const BfastMatrix2x2 &a, const BfastMatrix2x2 &b);
// These are component-wise operations.

inline BfastMatrix2x2 operator*(const BfastMatrix2x2 &a, const BfastMatrix2x2 &b); 
inline BfastVector2   operator*(const BfastMatrix2x2 &a, const BfastVector2   &b);
inline BfastVector2   operator*(const BfastVector2   &a, const BfastMatrix2x2 &b);
// Matrix multiplication.  Vector returning routines treat the vector
// argument like a row or column vector as appropriate.

//-------------------------------------------------------------------
//-------------------------------------------------------------------
// Global funtions for class BfastMatrix2x2

inline double determinant(const BfastMatrix2x2 &a);

inline double trace(const BfastMatrix2x2 &a);

inline BfastMatrix2x2 transpose(const BfastMatrix2x2 &a);
/*  */ BfastMatrix2x2   inverse(const BfastMatrix2x2 &a);

inline BfastMatrix2x2 diagonal(const BfastVector2 &a);

void BfastSymetricEigenDecomp(const BfastMatrix2x2 &a,
			   /* */ BfastVector2   &vals,
			   /* */ BfastMatrix2x2 &vecs);
// Given a SYMETRIC matrix, a, it returns the Eigen Values in vals,
// and the Eigen Vectors as the COLUMNS of vecs.  The vectors are
// normalized.  The values/vectors are not sorted in any order.


inline BfastMatrix2x2 BfastMatrixProduct(const BfastVector2 a,const BfastVector2 b);
// Returns the matrix given by a * aT  [ m_ij = a_i * b_j ]

inline BfastMatrix2x2 BfastEigenMatrix(const BfastVector2 a);
// Returns BfastMatrixProduct(a,a)/|a| [ m_ij = a_i * a_j / (a_k * a_k)].
// Note that the (only) eigen vector of the matrix is a, and the
// matrix is symetric.  (Involves a call to sqrt.)



std::istream &operator>>(std::istream &strm,      BfastMatrix2x2 &m);
std::ostream &operator<<(std::ostream &strm,const BfastMatrix2x2 &m);


//-------------------------------------------------------------------
//-------------------------------------------------------------------
//-------------------------------------------------------------------
//-------------------------------------------------------------------
// Inline implementation below:::

#ifndef DB_CHECK
#ifdef DEBUG
#define DB_CHECK( C ) { if ( ! (C) ) { abort(); } }
#else
#define DB_CHECK( C ) { }
#endif
#endif

#include <cmath>
#include <cstdlib>
#include "bfastVector.H"


//-------------------------------------------------------------------
//-------------------------------------------------------------------
//-------------------------------------------------------------------
// Inline implementation for Methods from BfastMatrix3x3


inline BfastMatrix3x3::BfastMatrix3x3() {
  data[0][0] = 1;
  data[0][1] = 0;
  data[0][2] = 0;

  data[1][0] = 0;
  data[1][1] = 1;
  data[1][2] = 0;

  data[2][0] = 0;
  data[2][1] = 0;
  data[2][2] = 1;  
}

inline BfastMatrix3x3::BfastMatrix3x3(double x) {
  data[0][0] = x;
  data[0][1] = x;
  data[0][2] = x;

  data[1][0] = x;
  data[1][1] = x;
  data[1][2] = x;

  data[2][0] = x;
  data[2][1] = x;
  data[2][2] = x;
}

inline BfastMatrix3x3::BfastMatrix3x3(const BfastMatrix3x3 &that) {
  data[0][0] = that.data[0][0];
  data[0][1] = that.data[0][1];
  data[0][2] = that.data[0][2];

  data[1][0] = that.data[1][0];
  data[1][1] = that.data[1][1];
  data[1][2] = that.data[1][2];

  data[2][0] = that.data[2][0];
  data[2][1] = that.data[2][1];
  data[2][2] = that.data[2][2];
};

inline BfastMatrix3x3::BfastMatrix3x3(double xx, double xy, double xz,
				double yx, double yy, double yz,
				double zx, double zy, double zz) {
  data[0][0] = xx;
  data[0][1] = xy;
  data[0][2] = xz;

  data[1][0] = yx;
  data[1][1] = yy;
  data[1][2] = yz;

  data[2][0] = zx;
  data[2][1] = zy;
  data[2][2] = zz;
}

inline BfastMatrix3x3::BfastMatrix3x3(const double data[3][3], BfastMatrix3x3::ordering o) {
  switch (o) {
  case (BfastMatrix3x3::row):
    this->data[0][0] = data[0][0];
    this->data[0][1] = data[0][1];
    this->data[0][2] = data[0][2];

    this->data[1][0] = data[1][0];
    this->data[1][1] = data[1][1];
    this->data[1][2] = data[1][2];

    this->data[2][0] = data[2][0];
    this->data[2][1] = data[2][1];
    this->data[2][2] = data[2][2];
    break;
  case (BfastMatrix3x3::col):
    this->data[0][0] = data[0][0];
    this->data[0][1] = data[1][0];
    this->data[0][2] = data[2][0];

    this->data[1][0] = data[0][1];
    this->data[1][1] = data[1][1];
    this->data[1][2] = data[2][1];

    this->data[2][0] = data[0][2];
    this->data[2][1] = data[1][2];
    this->data[2][2] = data[2][2];
    break;
  default:
    abort();
  }
}


inline BfastMatrix3x3::BfastMatrix3x3(const BfastVector3 data[3], BfastMatrix3x3::ordering o) {
  switch (o) {
  case (BfastMatrix3x3::row):
    this->data[0][0] = data[0][0];
    this->data[0][1] = data[0][1];
    this->data[0][2] = data[0][2];

    this->data[1][0] = data[1][0];
    this->data[1][1] = data[1][1];
    this->data[1][2] = data[1][2];

    this->data[2][0] = data[2][0];
    this->data[2][1] = data[2][1];
    this->data[2][2] = data[2][2];
    break;
  case (BfastMatrix3x3::col):
    this->data[0][0] = data[0][0];
    this->data[0][1] = data[1][0];
    this->data[0][2] = data[2][0];

    this->data[1][0] = data[0][1];
    this->data[1][1] = data[1][1];
    this->data[1][2] = data[2][1];

    this->data[2][0] = data[0][2];
    this->data[2][1] = data[1][2];
    this->data[2][2] = data[2][2];
    break;
  default:
    abort();
  }
}


inline BfastMatrix3x3::BfastMatrix3x3(const BfastVector3 a,
				const BfastVector3 b,
				const BfastVector3 c, BfastMatrix3x3::ordering o) {
  switch (o) {
  case (BfastMatrix3x3::row):
    this->data[0][0] = a[0];
    this->data[0][1] = a[1];
    this->data[0][2] = a[2];

    this->data[1][0] = b[0];
    this->data[1][1] = b[1];
    this->data[1][2] = b[2];

    this->data[2][0] = c[0];
    this->data[2][1] = c[1];
    this->data[2][2] = c[2];
    break;
  case (BfastMatrix3x3::col):
    this->data[0][0] = a[0];
    this->data[0][1] = b[0];
    this->data[0][2] = c[0];

    this->data[1][0] = a[1];
    this->data[1][1] = b[1];
    this->data[1][2] = c[1];

    this->data[2][0] = a[2];
    this->data[2][1] = b[2];
    this->data[2][2] = c[2];
    break;
  default:
    abort();
  }
}


//-------------------------------------------------------------------


inline double &BfastMatrix3x3::operator()(unsigned int r,unsigned int c) {
   DB_CHECK(r<3);
   DB_CHECK(c<3);
   return data[r][c];
}

inline double BfastMatrix3x3::operator()(unsigned int r,unsigned int c) const { 
   DB_CHECK(r<3);
   DB_CHECK(c<3);
   return data[r][c];
}


//-------------------------------------------------------------------


inline BfastMatrix3x3 &BfastMatrix3x3::set(const double d) {
  return (*this)=d;
}

inline BfastMatrix3x3 &BfastMatrix3x3::operator=(const double d) {
  data[0][0] = d;
  data[0][1] = d;
  data[0][2] = d;

  data[1][0] = d;
  data[1][1] = d;
  data[1][2] = d;

  data[2][0] = d;
  data[2][1] = d;
  data[2][2] = d;
  return (*this);
};


//-------------------------------------------------------------------

inline BfastMatrix3x3 &BfastMatrix3x3::set(const BfastMatrix3x3 &that) {
  return (*this)=that;
}

inline BfastMatrix3x3 &BfastMatrix3x3::operator=(const BfastMatrix3x3 &that) {
  data[0][0] = that.data[0][0];
  data[0][1] = that.data[0][1];
  data[0][2] = that.data[0][2];

  data[1][0] = that.data[1][0];
  data[1][1] = that.data[1][1];
  data[1][2] = that.data[1][2];

  data[2][0] = that.data[2][0];
  data[2][1] = that.data[2][1];
  data[2][2] = that.data[2][2];
  return (*this);
};



//-------------------------------------------------------------------

inline int BfastMatrix3x3::operator==(double d) const {
  return  ( (data[0][0] == d) &&
	    (data[0][1] == d) &&
	    (data[0][2] == d) &&
	    (data[1][0] == d) &&
	    (data[1][1] == d) &&
	    (data[1][2] == d) &&
	    (data[2][0] == d) &&
	    (data[2][1] == d) &&
	    (data[2][2] == d) );
}

inline int BfastMatrix3x3::operator!=(double d) const {
  return  ( (data[0][0] != d) ||
	    (data[0][1] != d) ||
	    (data[0][2] != d) ||
	    (data[1][0] != d) ||
	    (data[1][1] != d) ||
	    (data[1][2] != d) ||
	    (data[2][0] != d) ||
	    (data[2][1] != d) ||
	    (data[2][2] != d) );
}
  
inline int BfastMatrix3x3::operator==(const BfastMatrix3x3 &that)const {
  return ( (data[0][0] == that.data[0][0]) &&
	   (data[0][1] == that.data[0][1]) &&
	   (data[0][2] == that.data[0][2]) &&
	   (data[1][0] == that.data[1][0]) &&
	   (data[1][1] == that.data[1][1]) &&
	   (data[1][2] == that.data[1][2]) &&
	   (data[2][0] == that.data[2][0]) &&
	   (data[2][1] == that.data[2][1]) &&
	   (data[2][2] == that.data[2][2]) );
}

inline int BfastMatrix3x3::operator!=(const BfastMatrix3x3 &that)const {
  return ( (data[0][0] != that.data[0][0]) ||
	   (data[0][1] != that.data[0][1]) ||
	   (data[0][2] != that.data[0][2]) ||
	   (data[1][0] != that.data[1][0]) ||
	   (data[1][1] != that.data[1][1]) ||
	   (data[1][2] != that.data[1][2]) ||
	   (data[2][0] != that.data[2][0]) ||
	   (data[2][1] != that.data[2][1]) ||
	   (data[2][2] != that.data[2][2]) );
}


//-------------------------------------------------------------------

inline BfastMatrix3x3 &BfastMatrix3x3::operator+=(double d) {
  data[0][0] += d; data[1][0] += d; data[2][0] += d;
  data[0][1] += d; data[1][1] += d; data[2][1] += d;
  data[0][2] += d; data[1][2] += d; data[2][2] += d;
  return (*this);
}

inline BfastMatrix3x3 &BfastMatrix3x3::operator-=(double d) {
  data[0][0] -= d; data[1][0] -= d; data[2][0] -= d;
  data[0][1] -= d; data[1][1] -= d; data[2][1] -= d;
  data[0][2] -= d; data[1][2] -= d; data[2][2] -= d;
  return (*this);
}

inline BfastMatrix3x3 &BfastMatrix3x3::operator*=(double d) {
  data[0][0] *= d; data[1][0] *= d; data[2][0] *= d;
  data[0][1] *= d; data[1][1] *= d; data[2][1] *= d;
  data[0][2] *= d; data[1][2] *= d; data[2][2] *= d;
  return (*this);
}

inline BfastMatrix3x3 &BfastMatrix3x3::operator/=(double d) {
  data[0][0] /= d; data[1][0] /= d; data[2][0] /= d;
  data[0][1] /= d; data[1][1] /= d; data[2][1] /= d;
  data[0][2] /= d; data[1][2] /= d; data[2][2] /= d;
  return (*this);
}

//-------------------------------------------------------------------

inline BfastMatrix3x3 &BfastMatrix3x3::operator+=(const BfastMatrix3x3 &that) {
  return componentAdd(that);
}
  
inline BfastMatrix3x3 &BfastMatrix3x3::operator-=(const BfastMatrix3x3 &that) {
  return componentSub(that);
}

//-------------------------------------------------------------------

inline BfastMatrix3x3 &BfastMatrix3x3::componentAdd (const BfastMatrix3x3 &that) {
  data[0][0] += that(0,0); data[1][0] += that(1,0); data[2][0] += that(2,0);
  data[0][1] += that(0,1); data[1][1] += that(1,1); data[2][1] += that(2,1);
  data[0][2] += that(0,2); data[1][2] += that(1,2); data[2][2] += that(2,2);
  return (*this);
}

inline BfastMatrix3x3 &BfastMatrix3x3::componentSub (const BfastMatrix3x3 &that) {
  data[0][0] -= that(0,0); data[1][0] -= that(1,0); data[2][0] -= that(2,0);
  data[0][1] -= that(0,1); data[1][1] -= that(1,1); data[2][1] -= that(2,1);
  data[0][2] -= that(0,2); data[1][2] -= that(1,2); data[2][2] -= that(2,2);
  return (*this);
}

inline BfastMatrix3x3 &BfastMatrix3x3::componentMult(const BfastMatrix3x3 &that) {
  data[0][0] *= that(0,0); data[1][0] *= that(1,0); data[2][0] *= that(2,0);
  data[0][1] *= that(0,1); data[1][1] *= that(1,1); data[2][1] *= that(2,1);
  data[0][2] *= that(0,2); data[1][2] *= that(1,2); data[2][2] *= that(2,2);
  return (*this);
}

inline BfastMatrix3x3 &BfastMatrix3x3::componentDiv (const BfastMatrix3x3 &that) {
  data[0][0] /= that(0,0); data[1][0] /= that(1,0); data[2][0] /= that(2,0);
  data[0][1] /= that(0,1); data[1][1] /= that(1,1); data[2][1] /= that(2,1);
  data[0][2] /= that(0,2); data[1][2] /= that(1,2); data[2][2] /= that(2,2);
  return (*this);
}


//-------------------------------------------------------------------

inline BfastMatrix3x3 &BfastMatrix3x3::setIdentity() {
  data[0][0] = 1; data[0][1] = 0; data[0][2] = 0;
  data[1][0] = 0; data[1][1] = 1; data[1][2] = 0;
  data[2][0] = 0; data[2][1] = 0; data[2][2] = 1;
  return (*this);
};

inline BfastMatrix3x3 &BfastMatrix3x3::inplaceTranspose() {
  double tmp;
  tmp = data[0][1]; data[0][1] = data[1][0]; data[1][0] = tmp;
  tmp = data[0][2]; data[0][2] = data[2][0]; data[2][0] = tmp;
  tmp = data[1][2]; data[1][2] = data[2][1]; data[2][1] = tmp;
  return (*this);
}

//-------------------------------------------------------------------

inline unsigned int BfastMatrix3x3::cycleAxis(unsigned int axis, int direction) {
  return BfastVector3::cycleAxis(axis,direction);
}



//-------------------------------------------------------------------
//-------------------------------------------------------------------
//-------------------------------------------------------------------
// Inline implementation for Methods from BfastMatrix2x2


inline BfastMatrix2x2::BfastMatrix2x2() {
  data[0][0] = 1;
  data[0][1] = 0;

  data[1][0] = 0;
  data[1][1] = 1;
}

inline BfastMatrix2x2::BfastMatrix2x2(double x) {
  data[0][0] = x;
  data[0][1] = x;

  data[1][0] = x;
  data[1][1] = x;
}


inline BfastMatrix2x2::BfastMatrix2x2(double xx, double xy, 
				double yx, double yy) {
  data[0][0] = xx;
  data[0][1] = xy;
  data[1][0] = yx;
  data[1][1] = yy;
};



inline BfastMatrix2x2::BfastMatrix2x2(const BfastMatrix2x2 &that) {
  data[0][0] = that.data[0][0];
  data[0][1] = that.data[0][1];

  data[1][0] = that.data[1][0];
  data[1][1] = that.data[1][1];
};

inline BfastMatrix2x2::BfastMatrix2x2(const double data[2][2], BfastMatrix2x2::ordering o) {
  switch (o) {
  case (BfastMatrix2x2::row):
    this->data[0][0] = data[0][0];
    this->data[0][1] = data[0][1];

    this->data[1][0] = data[1][0];
    this->data[1][1] = data[1][1];
    break;
  case (BfastMatrix2x2::col):
    this->data[0][0] = data[0][0];
    this->data[0][1] = data[1][0];

    this->data[1][0] = data[0][1];
    this->data[1][1] = data[1][1];
  default:
    abort();
  }
}


inline BfastMatrix2x2::BfastMatrix2x2(const BfastVector2 data[2], BfastMatrix2x2::ordering o) {
  switch (o) {
  case (BfastMatrix2x2::row):
    this->data[0][0] = data[0][0];
    this->data[0][1] = data[0][1];

    this->data[1][0] = data[1][0];
    this->data[1][1] = data[1][1];
    break;
  case (BfastMatrix2x2::col):
    this->data[0][0] = data[0][0];
    this->data[0][1] = data[1][0];

    this->data[1][0] = data[0][1];
    this->data[1][1] = data[1][1];
    break;
  default:
    abort();
  }
}


inline BfastMatrix2x2::BfastMatrix2x2(const BfastVector2 a,
				const BfastVector2 b, BfastMatrix2x2::ordering o) {
  switch (o) {
  case (BfastMatrix2x2::row):
    this->data[0][0] = a[0];
    this->data[0][1] = a[1];

    this->data[1][0] = b[0];
    this->data[1][1] = b[1];
    break;
  case (BfastMatrix2x2::col):
    this->data[0][0] = a[0];
    this->data[0][1] = b[0];

    this->data[1][0] = a[1];
    this->data[1][1] = b[1];
    break;
  default:
    abort();
  }
}


//-------------------------------------------------------------------


inline double &BfastMatrix2x2::operator()(unsigned int r,unsigned int c) {
   DB_CHECK(r<2);
   DB_CHECK(c<2);
   return data[r][c];
}

inline double BfastMatrix2x2::operator()(unsigned int r,unsigned int c) const { 
   DB_CHECK(r<2);
   DB_CHECK(c<2);
   return data[r][c];
}


//-------------------------------------------------------------------


inline BfastMatrix2x2 &BfastMatrix2x2::set(const double d) {
  return (*this)=d;
}

inline BfastMatrix2x2 &BfastMatrix2x2::operator=(const double d) {
  data[0][0] = d;
  data[0][1] = d;

  data[1][0] = d;
  data[1][1] = d;
  return (*this);
};


//-------------------------------------------------------------------

inline BfastMatrix2x2 &BfastMatrix2x2::set(const BfastMatrix2x2 &that) {
  return (*this)=that;
}

inline BfastMatrix2x2 &BfastMatrix2x2::operator=(const BfastMatrix2x2 &that) {
  data[0][0] = that.data[0][0];
  data[0][1] = that.data[0][1];

  data[1][0] = that.data[1][0];
  data[1][1] = that.data[1][1];
  return (*this);
};



//-------------------------------------------------------------------

inline int BfastMatrix2x2::operator==(double d) const {
  return  ( (data[0][0] == d) &&
	    (data[0][1] == d) &&
	    (data[1][0] == d) &&
	    (data[1][1] == d) );
}

inline int BfastMatrix2x2::operator!=(double d) const {
  return  ( (data[0][0] != d) ||
	    (data[0][1] != d) ||
	    (data[1][0] != d) ||
	    (data[1][1] != d) );
}
  
inline int BfastMatrix2x2::operator==(const BfastMatrix2x2 &that)const {
  return ( (data[0][0] == that.data[0][0]) &&
	   (data[0][1] == that.data[0][1]) &&
	   (data[1][0] == that.data[1][0]) &&
	   (data[1][1] == that.data[1][1]) );
}

inline int BfastMatrix2x2::operator!=(const BfastMatrix2x2 &that)const {
  return ( (data[0][0] != that.data[0][0]) ||
	   (data[0][1] != that.data[0][1]) ||
	   (data[1][0] != that.data[1][0]) ||
	   (data[1][1] != that.data[1][1]) );
}


//-------------------------------------------------------------------

inline BfastMatrix2x2 &BfastMatrix2x2::operator+=(double d) {
  data[0][0] += d; data[1][0] += d; 
  data[0][1] += d; data[1][1] += d; 
  return (*this);
}

inline BfastMatrix2x2 &BfastMatrix2x2::operator-=(double d) {
  data[0][0] -= d; data[1][0] -= d; 
  data[0][1] -= d; data[1][1] -= d; 
  return (*this);
}

inline BfastMatrix2x2 &BfastMatrix2x2::operator*=(double d) {
  data[0][0] *= d; data[1][0] *= d; 
  data[0][1] *= d; data[1][1] *= d; 
  return (*this);
}

inline BfastMatrix2x2 &BfastMatrix2x2::operator/=(double d) {
  data[0][0] /= d; data[1][0] /= d; 
  data[0][1] /= d; data[1][1] /= d; 
  return (*this);
}

//-------------------------------------------------------------------

inline BfastMatrix2x2 &BfastMatrix2x2::operator+=(const BfastMatrix2x2 &that) {
  return componentAdd(that);
}
  
inline BfastMatrix2x2 &BfastMatrix2x2::operator-=(const BfastMatrix2x2 &that) {
  return componentSub(that);
}

//-------------------------------------------------------------------

inline BfastMatrix2x2 &BfastMatrix2x2::componentAdd (const BfastMatrix2x2 &that) {
  data[0][0] += that(0,0); data[1][0] += that(1,0); 
  data[0][1] += that(0,1); data[1][1] += that(1,1); 
  return (*this);
}

inline BfastMatrix2x2 &BfastMatrix2x2::componentSub (const BfastMatrix2x2 &that) {
  data[0][0] -= that(0,0); data[1][0] -= that(1,0); 
  data[0][1] -= that(0,1); data[1][1] -= that(1,1); 
  return (*this);
}

inline BfastMatrix2x2 &BfastMatrix2x2::componentMult(const BfastMatrix2x2 &that) {
  data[0][0] *= that(0,0); data[1][0] *= that(1,0); 
  data[0][1] *= that(0,1); data[1][1] *= that(1,1); 
  return (*this);
}

inline BfastMatrix2x2 &BfastMatrix2x2::componentDiv (const BfastMatrix2x2 &that) {
  data[0][0] /= that(0,0); data[1][0] /= that(1,0); 
  data[0][1] /= that(0,1); data[1][1] /= that(1,1); 
  return (*this);
}


//-------------------------------------------------------------------

inline BfastMatrix2x2 &BfastMatrix2x2::setIdentity() {
  data[0][0] = 1; data[0][1] = 0; 
  data[1][0] = 0; data[1][1] = 1; 
  return (*this);
};

inline BfastMatrix2x2 &BfastMatrix2x2::inplaceTranspose() {
  double tmp;
  tmp = data[0][1]; data[0][1] = data[1][0]; data[1][0] = tmp;
  return (*this);
}

//-------------------------------------------------------------------

inline unsigned int BfastMatrix2x2::cycleAxis(unsigned int axis, int direction) {
  return BfastVector2::cycleAxis(axis,direction);
}

//-------------------------------------------------------------------
//-------------------------------------------------------------------
//-------------------------------------------------------------------
// Inline implementation of global functions for BfastMatrix3x3

inline BfastMatrix3x3 operator+(const BfastMatrix3x3 &a,double b) {
  return (BfastMatrix3x3(a)+=b);
}

inline BfastMatrix3x3 operator-(const BfastMatrix3x3 &a,double b) {
  return (BfastMatrix3x3(a)-=b);
}

inline BfastMatrix3x3 operator*(const BfastMatrix3x3 &a,double b) {
  return (BfastMatrix3x3(a)*=b);
}
 
inline BfastMatrix3x3 operator/(const BfastMatrix3x3 &a,double b) {
  return (BfastMatrix3x3(a)/=b);  
}


//-------------------------------------------------------------------

inline BfastMatrix3x3 operator+(double a, const BfastMatrix3x3 &b) {
  return b+a;
}

inline BfastMatrix3x3 operator-(double a, const BfastMatrix3x3 &b) {
  return BfastMatrix3x3(a-b(0,0),a-b(0,1),a-b(0,2),
		     a-b(1,0),a-b(1,1),a-b(1,2),
		     a-b(2,0),a-b(2,1),a-b(2,2));
}

inline BfastMatrix3x3 operator*(double a, const BfastMatrix3x3 &b) {
  return b*a;
}
 
inline BfastMatrix3x3 operator/(double a, const BfastMatrix3x3 &b) {
  return BfastMatrix3x3(a/b(0,0),a/b(0,1),a/b(0,2),
		     a/b(1,0),a/b(1,1),a/b(1,2),
		     a/b(2,0),a/b(2,1),a/b(2,2));
}


//-------------------------------------------------------------------

inline BfastMatrix3x3 operator+(const BfastMatrix3x3 &a,const BfastMatrix3x3 &b) {
  return (BfastMatrix3x3(a)+=b);
}
 
inline BfastMatrix3x3 operator-(const BfastMatrix3x3 &a,const BfastMatrix3x3 &b) {
  return (BfastMatrix3x3(a)-=b);
}
  
//-------------------------------------------------------------------

inline BfastMatrix3x3 operator*(const BfastMatrix3x3 &a,const BfastMatrix3x3 &b) {
  BfastMatrix3x3 tmp(a);
  tmp.inplaceMultPost(b);
  return tmp;
}

inline BfastVector3 operator*(const BfastMatrix3x3 &a,const BfastVector3 &b) {
  return BfastVector3(b[0]*a(0,0) + b[1]*a(0,1) + b[2]*a(0,2),
		   b[0]*a(1,0) + b[1]*a(1,1) + b[2]*a(1,2),
		   b[0]*a(2,0) + b[1]*a(2,1) + b[2]*a(2,2));
}

inline BfastVector3 operator*(const BfastVector3 &a,const BfastMatrix3x3 &b) {
  return BfastVector3(a[0]*b(0,0) + a[1]*b(1,0) + a[2]*b(2,0),
		   a[0]*b(0,1) + a[1]*b(1,1) + a[2]*b(2,1),
		   a[0]*b(0,2) + a[1]*b(1,2) + a[2]*b(2,2));
}


//-------------------------------------------------------------------  

inline double determinant(const BfastMatrix3x3 &a) {
  return ( a(0,0) * ( a(1,1) * a(2,2) - a(1,2) * a(2,1) ) +
	   a(0,1) * ( a(1,2) * a(2,0) - a(1,0) * a(2,2) ) +
	   a(0,2) * ( a(1,0) * a(2,1) - a(1,1) * a(2,0) ) );
}


inline double trace(const BfastMatrix3x3 &a) {
  return ( a(0,0) + a(1,1) + a(2,2) );
}



inline BfastMatrix3x3 transpose(const BfastMatrix3x3 &a) {
  BfastMatrix3x3 tmp(a);
  tmp.inplaceTranspose();
  return tmp;
}
  

inline BfastMatrix3x3 diagonal(const BfastVector3 &a) {
  BfastMatrix3x3 tmp(0.0);
  tmp(0,0) = a(0);
  tmp(1,1) = a(1);
  tmp(2,2) = a(2);
  return tmp;
}
  



//-------------------------------------------------------------------

inline BfastMatrix3x3 BfastMatrixProduct(const BfastVector3 a,const BfastVector3 b) {
  return BfastMatrix3x3( a[0] * b[0] ,  a[0] * b[1] ,  a[0] * b[2] , 
		      a[1] * b[0] ,  a[1] * b[1] ,  a[1] * b[2] , 
		      a[2] * b[0] ,  a[2] * b[1] ,  a[2] * b[2] );
}
		     
inline BfastMatrix3x3 BfastEigenMatrix(const BfastVector3 a) {
  double m = mag(a);
  if (m==0) {
    return 0.0;
  }else{
    return BfastMatrixProduct(a,a*(1.0/m));
  }
}



//-------------------------------------------------------------------
//-------------------------------------------------------------------
//-------------------------------------------------------------------
// Inline implementation of global functions for BfastMatrix2x2

inline BfastMatrix2x2 operator+(const BfastMatrix2x2 &a,double b) {
  return (BfastMatrix2x2(a)+=b);
}

inline BfastMatrix2x2 operator-(const BfastMatrix2x2 &a,double b) {
  return (BfastMatrix2x2(a)-=b);
}

inline BfastMatrix2x2 operator*(const BfastMatrix2x2 &a,double b) {
  return (BfastMatrix2x2(a)*=b);
}
 
inline BfastMatrix2x2 operator/(const BfastMatrix2x2 &a,double b) {
  return (BfastMatrix2x2(a)/=b);  
}


//-------------------------------------------------------------------

inline BfastMatrix2x2 operator+(double a, const BfastMatrix2x2 &b) {
return b+a;
}

inline BfastMatrix2x2 operator-(double a, const BfastMatrix2x2 &b) {
  return BfastMatrix2x2(a-b(0,1),a-b(0,1),
		     a-b(1,1),a-b(1,1));
}

inline BfastMatrix2x2 operator*(double a, const BfastMatrix2x2 &b) {
  return b*a;
}
 
inline BfastMatrix2x2 operator/(double a, const BfastMatrix2x2 &b) {
  return BfastMatrix2x2(a/b(0,1),a/b(0,1),
		     a/b(1,1),a/b(1,1));
}


//-------------------------------------------------------------------

inline BfastMatrix2x2 operator+(const BfastMatrix2x2 &a,const BfastMatrix2x2 &b) {
  return (BfastMatrix2x2(a)+=b);
}
 
inline BfastMatrix2x2 operator-(const BfastMatrix2x2 &a,const BfastMatrix2x2 &b) {
  return (BfastMatrix2x2(a)-=b);
}
  
  
//-------------------------------------------------------------------

inline BfastMatrix2x2 operator*(const BfastMatrix2x2 &a,const BfastMatrix2x2 &b) {
  BfastMatrix2x2 tmp(a);
  tmp.inplaceMultPost(b);
  return tmp;
}

inline BfastVector2 operator*(const BfastMatrix2x2 &a,const BfastVector2 &b) {
  return BfastVector2(b[0]*a(0,0) + b[1]*a(0,1),
		   b[0]*a(1,0) + b[1]*a(1,1));
}

inline BfastVector2 operator*(const BfastVector2 &a,const BfastMatrix2x2 &b) {
  return BfastVector2(a[0]*b(0,0) + a[1]*b(1,0),
		   a[0]*b(0,1) + a[1]*b(1,1));
}


//-------------------------------------------------------------------  

inline double determinant(const BfastMatrix2x2 &a) {
  return ( a(0,0) * a(1,1) - a(0,1) * a(1,0) );
}

inline double trace(const BfastMatrix2x2 &a) {
  return ( a(0,0) + a(1,1) );
}



inline BfastMatrix2x2 transpose(const BfastMatrix2x2 &a) {
  BfastMatrix2x2 tmp(a);
  tmp.inplaceTranspose();
  return tmp;
}
 
inline BfastMatrix2x2 diagonal(const BfastVector2 &a) {
  BfastMatrix2x2 tmp(0.0);
  tmp(0,0) = a(0);
  tmp(1,1) = a(1);
  return tmp;
}
  

//-------------------------------------------------------------------

inline BfastMatrix2x2 BfastMatrixProduct(const BfastVector2 a,const BfastVector2 b) {
  return BfastMatrix2x2( a[0] * b[0] ,  a[0] * b[1] ,
		      a[1] * b[0] ,  a[1] * b[1] );
}
		     
inline BfastMatrix2x2 BfastEigenMatrix(const BfastVector2 a) {
  double m = mag(a);
  if (m==0) {
    return 0.0;
  }else{
    return BfastMatrixProduct(a,a*(1.0/m));
  }
}


//-------------------------------------------------------------------
//-------------------------------------------------------------------
//-------------------------------------------------------------------

#endif
