//-------------------------------------------------------------------
//-------------------------------------------------------------------
//
// Useful Stuff
//   -- Code File
//  
// Primary Author: Adam Bargteil (adamb@cs.berkeley.edu)
// 
//-------------------------------------------------------------------
//-------------------------------------------------------------------
//
// Copyright (c) 2003-2005, Regents of the University of California.  All
// rights reserved.
//
// This software is part of the Berkeley Fluid Animation & Simulation
// Toolkit.  The name "Berkeley Fluid Animation & Simulation Toolkit" is
// a trademark of the Regents of the University of California.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//   Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//
//  Redistributions in binary form must reproduce the above copyright
//  notice, this list of conditions and the following disclaimer in the
//  documentation and/or other materials provided with the distribution.
//
//  Redistributions in binary form as an executable program, or part of
//  an executable program must, when run for the first time by a given
//  user, prominently display the above copyright notice, this list of
//  conditions and the following disclaimer.
//
//  Neither the name of the University of California, Berkeley nor the
//  names of its contributors may be used to endorse or promote products
//  derived from this software without specific prior written
//  permission.
//
//  ** Animations, still images, or other works created using this
//  ** software must clearly indicate in the list of credits that this
//  ** software was used.  The software shall be referred to as the
//  ** Berkeley Fluid Animation & Simulation Toolkit.  If the software
//  ** is included, either in whole or in part, within another software
//  ** package, credit must still be given to the Berkeley Fluid
//  ** Animation & Simulation Toolkit.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
//-------------------------------------------------------------------
//-------------------------------------------------------------------

#include "bfastUtil.H"
#include <iostream>
#include <fstream>

using namespace std;

inline static
istream &eatChar(char c,istream &buf) {
  char r;
  buf >> r;
  if (r!=c) {
    buf.clear(buf.rdstate() | ios::failbit);
  }
  return buf;
}

inline static
istream &eatStr(const char *s,istream &buf) {
  while (*s != '\0') {
    eatChar(*s,buf);
    s++;
  }
  return buf;
}


BfastReal pow2(int p) {
  switch (p) {
	case -20: return 9.53674e-07;
	case -19: return 1.90735e-06;
	case -18: return 3.8147e-06;
	case -17: return 7.62939e-06;
	case -16: return 1.52588e-05;
	case -15: return 3.05176e-05;
	case -14: return 6.10352e-05;
	case -13: return 0.0001220703125;
	case -12: return 0.000244140625;
	case -11: return 0.00048828125;
	case -10: return 0.0009765625;
	case -9: return 0.001953125;
	case -8: return 0.00390625;
	case -7: return 0.0078125;
	case -6: return 0.015625;
	case -5: return 0.03125;
	case -4: return 0.0625;
	case -3: return 0.125;
	case -2: return 0.25;
	case -1: return 0.5;
	case 0: return 1;
	case 1: return 2;
	case 2: return 4;
	case 3: return 8;
	case 4: return 16;
	case 5: return 32;
	case 6: return 64;
	case 7: return 128;
	case 8: return 256;
	case 9: return 512;
	case 10: return 1024;
	case 11: return 2048;
	case 12: return 4096;
	case 13: return 8192;
	case 14: return 16384;
	case 15: return 32768;
	case 16: return 65536;
	case 17: return 131072;
	case 18: return 262144;
	case 19: return 524288;
	case 20: return 1048576;
  default:
    BfastReal ret = 1;
    if (abs(p) == p)
      for (int i=0; i<abs(p); i++)
				ret *= 2.0;
    else
      for (int i=0; i<abs(p); i++)
				ret /= 2.0;
    return ret;
  }
}


istream &operator>>(std::istream &strm, std::vector<int> &l) {
  unsigned int j,s = 0;
	std::vector<int>::iterator i;
  ios::fmtflags orgFlags = strm.setf(ios::skipws);

  eatStr("[",strm);
  strm >> s;
  eatStr(":",strm);
  
  if (strm.good()) {
    l.resize(s);
    for(i=l.begin(),j=0;i!=l.end();i++,j++) {
      strm >> (*i);
      if (j != (s-1)) eatStr(",",strm);
    }
  }
  eatChar(']',strm);
  strm.flags(orgFlags);
  return strm;
}

ostream &operator<<(ostream &strm,const std::vector<int> &l) {
  strm << "[";
  strm << l.size();
  strm << ":";
  unsigned int j;
	std::vector<int>::const_iterator i;
  for(i=l.begin(),j=0;i!=l.end();i++,j++) {
    strm << *i;
    if (j != (l.size()-1)) strm << ",";
  }
  strm << "]";
  return strm;
}

void bfast::read(ifstream &in, std::vector<int> &l) {
	unsigned int s;
	std::vector<int>::iterator i;
	in.read((char *)&s, sizeof(int));
	l.resize(s);
	for (i=l.begin(); i!=l.end(); i++) {
		in.read((char*)(&(*i)), sizeof(int));
	}
}

void bfast::write(ofstream &out, const std::vector<int> &l) {
	unsigned int s=l.size();
	std::vector<int>::const_iterator i;
	out.write((char*)&s, sizeof(int));
	for (i=l.begin(); i!=l.end(); i++) {
		out.write((char*) (&(*i)), sizeof(int));
	}
}

void randomize(std::vector<int> &list) {
	int i=list.size(), j, tmp;

	for (i=0; i>0;) {
		j = rand()%(i--);
		tmp = list[i];
		list[i] = list[j];
		list[j] = tmp;
	}
		
}

//////////////////////////////////////////////////////////////////
// BfastRealList
//////////////////////////////////////////////////////////////////
istream &operator>>(std::istream &strm, std::vector<BfastReal> &l) {
	unsigned int j,s=0;
	std::vector<BfastReal>::iterator i;
  ios::fmtflags orgFlags = strm.setf(ios::skipws);
	
  eatStr("[",strm);
  strm >> s;
  eatStr(":",strm);
  
  if (strm.good()) {
    l.resize(s);
    for(i=l.begin(),j=0;i!=l.end();i++,j++) {
      strm >> *i;
      if (j != (s-1)) eatStr(",",strm);
    }
  }
  eatChar(']',strm);
  strm.flags(orgFlags);
  return strm;
}

ostream &operator<<(ostream &strm,const std::vector<BfastReal> &l) {
  unsigned int j;
	std::vector<BfastReal>::const_iterator i;
  strm << "[";
  strm << l.size();
  strm << ":";
  for(i=l.begin(),j=0;i!=l.end();i++,j++) {
    strm << *i;
    if (j != (l.size()-1)) strm << ",";
  }
  strm << "]";
  return strm;
}

void bfast::read(ifstream &in, std::vector<BfastReal> &l) {
	unsigned int s;
	std::vector<BfastReal>::iterator i;
	in.read((char *)&s, sizeof(int));
	l.resize(s);
	for (i=l.begin(); i!=l.end(); i++) {
		in.read((char*)(&(*i)), sizeof(BfastReal));
	}
}

void bfast::write(ofstream &out, const std::vector<BfastReal> &l) {
	unsigned int s=l.size();
	std::vector<BfastReal>::const_iterator i;
	out.write((char *)&s, sizeof(int));
	for (i=l.begin(); i!=l.end(); i++) {
		out.write((char*) (&(*i)), sizeof(BfastReal));
	}
}


//////////////////////////////////////////////////////////////////
// BfastTri
//////////////////////////////////////////////////////////////////
std::istream &operator>>(std::istream &strm, BfastTri &t) {
  ios::fmtflags orgFlags = strm.setf(ios::skipws);
  eatStr("[",strm);
  strm >> t[0];
  eatChar(',',strm);
  strm >> t[1];
  eatChar(',',strm);
  strm >> t[2];
  eatChar(']',strm);
  strm.flags(orgFlags);
  return strm;
}

std::ostream &operator<<(std::ostream &strm, const BfastTri &t) {
  strm << "[";
  strm << t[0];
  strm << ",";
  strm << t[1];
  strm << ",";
  strm << t[2];
  strm << "]";
  return strm;
}
void bfast::read(ifstream &in, BfastTri &t) {
	in.read((char*) &(t[0]), sizeof(int));
	in.read((char*) &(t[1]), sizeof(int));
	in.read((char*) &(t[2]), sizeof(int));
}
void bfast::write(ofstream &out, const BfastTri &_t) {
	BfastTri t(_t);
	out.write((char*) &(t[0]), sizeof(int));
	out.write((char*) &(t[1]), sizeof(int));
	out.write((char*) &(t[2]), sizeof(int));
}

//////////////////////////////////////////////////////////////////
// BfastTriList
//////////////////////////////////////////////////////////////////
istream &operator>>(std::istream &strm, std::vector<BfastTri> &l) {
  unsigned int j,s = 0;
	std::vector<BfastTri>::iterator i;
  ios::fmtflags orgFlags = strm.setf(ios::skipws);

  eatStr("[",strm);
  strm >> s;
  eatStr(":",strm);
  
  if (strm.good()) {
    l.resize(s);
    for(i=l.begin(),j=0;i!=l.end();i++,j++) {
      strm >> *i;
      if (j != (s-1)) eatStr(",",strm);
    }
  }
  eatChar(']',strm);
  strm.flags(orgFlags);
  return strm;
}

ostream &operator<<(ostream &strm, const std::vector<BfastTri> &l) {
  unsigned int j;
	std::vector<BfastTri>::const_iterator i;
  strm << "[";
  strm << l.size();
  strm << ":";
  for(i=l.begin(),j=0;i!=l.end();i++,j++) {
    strm << *i;
    if (j != (l.size()-1)) strm << ",";
  }
  strm << "]";
  return strm;
}

void bfast::read(ifstream &in, std::vector<BfastTri> &l) {
	unsigned int s;
	std::vector<BfastTri>::iterator i;
	in.read((char *)&s, sizeof(int));
	l.resize(s);
	for (i=l.begin(); i!=l.end(); i++) {
		bfast::read(in, *i);
	}
}

void bfast::write(ofstream &out, const std::vector<BfastTri> &l) {
	unsigned int s=l.size();
	std::vector<BfastTri>::const_iterator i;
	out.write((char*)&s, sizeof(int));
	for (i=l.begin(); i!=l.end(); i++) {
		bfast::write(out, *i);
	}
}

#ifdef INVENTOR_DEFINED
SoIndexedFaceSet *bfast::faceSet(const std::vector<BfastTri> &l) {
	std::vector<BfastTri>::const_iterator i;
	SoIndexedFaceSet *faces = new SoIndexedFaceSet();
	faces->ref();
	faces->coordIndex.setNum(4*l.size());
	int32_t *ptr = faces->coordIndex.startEditing();
	for (i=l.begin(); i!=l.end(); i++) {
		*(ptr++) = (*i)[0];
		*(ptr++) = (*i)[1];
		*(ptr++) = (*i)[2];
		*(ptr++) = -1;
	}
	faces->coordIndex.finishEditing();
	faces->unrefNoDelete();
	return faces;
}
#endif


//////////////////////////////////////////////////////////////////
// BfastVecList
//////////////////////////////////////////////////////////////////
#ifdef INVENTOR_DEFINED
SoCoordinate3 *bfast::coord(const std::vector<BfastVector3> &l) {
	std::vector<BfastVector3>::const_iterator i;
	SoCoordinate3 *coords = new SoCoordinate3();
	coords->ref();

	if (coords->point.getNum() != (int)l.size())
		coords->point.setNum(l.size());

	SbVec3f *ptr = coords->point.startEditing();
	for (i=l.begin(); i!=l.end(); i++) 
		(ptr++)->setValue(float((*i)[0]), float((*i)[1]), float((*i)[2]));
	coords->point.finishEditing();
	coords->unrefNoDelete();
	return coords;
}
#endif

istream &operator>>(std::istream &strm, std::vector<BfastVector3> &l) {
  unsigned int j,s = 0;
	std::vector<BfastVector3>::iterator i;
  ios::fmtflags orgFlags = strm.setf(ios::skipws);

  eatStr("[",strm);
  strm >> s;
  eatStr(":",strm);
  
  if (strm.good()) {
    l.resize(s);
    for(i=l.begin(),j=0;i!=l.end();i++,j++) {
      strm >> *i;
      if (j != (s-1)) eatStr(",",strm);
    }
  }
  eatChar(']',strm);
  strm.flags(orgFlags);
  return strm;
}

ostream &operator<<(ostream &strm,const std::vector<BfastVector3> &l) {
  unsigned int j;
	std::vector<BfastVector3>::const_iterator i;
  strm << "[";
  strm << l.size();
  strm << ":";
  for(i=l.begin(),j=0;i!=l.end();i++,j++) {
    strm << *i;
    if (j != (l.size()-1)) strm << ",";
  }
  strm << "]";
  return strm;
}

void bfast::read(ifstream &in, std::vector<BfastVector3> &l) {
	int s=0;
	std::vector<BfastVector3>::iterator i;
	in.read((char*)&s, sizeof(int));
	l.resize(s);
	for (i=l.begin(); i!=l.end(); i++) {
		in.read((char*)&((*i)[0]), sizeof(BfastReal));
		in.read((char*)&((*i)[1]), sizeof(BfastReal));
		in.read((char*)&((*i)[2]), sizeof(BfastReal));
	}
}

void bfast::write(ofstream &out, const std::vector<BfastVector3> &l) {
	int s=l.size();
	std::vector<BfastVector3>::const_iterator i;
	out.write((char*)&s, sizeof(int));
	for (i=l.begin(); i!=l.end(); i++) {
		BfastVector3 x(*i);
		out.write((char*)&(x[0]), sizeof(BfastReal));
		out.write((char*)&(x[1]), sizeof(BfastReal));
		out.write((char*)&(x[2]), sizeof(BfastReal));
	}
}

//////////////////////////////////////////////////////////////////
// BfastInt3List
//////////////////////////////////////////////////////////////////
istream &operator>>(std::istream &strm, BfastInt3 &v) {
	ios::fmtflags orgFlags = strm.setf(ios::skipws);
  eatChar('[',strm);
  strm >> v[0];
  eatChar(',',strm);
  strm >> v[1];
  eatChar(',',strm);
  strm >> v[2];
  eatChar(']',strm);
  strm.flags(orgFlags);
  return strm;
}
  
ostream &operator<<(ostream &strm,const BfastInt3 &v) {
  strm << "[";
	strm << v[0]; strm << ",";
	strm << v[1]; strm << ",";
	strm << v[2]; strm << "]";
  return strm;
}

istream &operator>>(std::istream &strm, std::vector<BfastInt3> &l) {
  unsigned int j,s = 0;
	std::vector<BfastInt3>::iterator i;
  ios::fmtflags orgFlags = strm.setf(ios::skipws);

  eatStr("[",strm);
  strm >> s;
  eatStr(":",strm);
  
  if (strm.good()) {
    l.resize(s);
    for(i=l.begin(),j=0;i!=l.end();i++,j++) {
      strm >> *i;
      if (j != (s-1)) eatStr(",",strm);
    }
  }
  eatChar(']',strm);
  strm.flags(orgFlags);
  return strm;
}

ostream &operator<<(ostream &strm,const std::vector<BfastInt3> &l) {
  unsigned int j;
	std::vector<BfastInt3>::const_iterator i;
  strm << "[";
  strm << l.size();
  strm << ":";
  for(i=l.begin(),j=0;i!=l.end();i++,j++) {
    strm << *i;
    if (j != (l.size()-1)) strm << ",";
  }
  strm << "]";
  return strm;
}

void bfast::read(ifstream &in, std::vector<BfastInt3> &l) {
	int s;
	std::vector<BfastInt3>::iterator i;
	in.read((char*)&s, sizeof(int));
	l.resize(s);
	for (i=l.begin(); i<l.end(); i++) {
		in.read((char*)&((*i)[0]), sizeof(int));
		in.read((char*)&((*i)[1]), sizeof(int));
		in.read((char*)&((*i)[2]), sizeof(int));
	}
}

void bfast::write(ofstream &out, const std::vector<BfastInt3> &l) {
	int s=l.size();
	std::vector<BfastInt3>::const_iterator i;
	out.write((char*)&s, sizeof(int));
	for (i=l.begin(); i<l.end(); i++) {
		BfastInt3 x(*i);
		out.write((char*)&(x[0]), sizeof(int));
		out.write((char*)&(x[1]), sizeof(int));
		out.write((char*)&(x[2]), sizeof(int));
	}
}

void computeNormals(const std::vector<BfastVector3> &meshPts, const std::vector<BfastTri> &triangles, 
										std::vector<BfastVector3> &vertexNormals, std::vector<BfastVector3> &faceNormals) {
	std::vector<BfastVector3>::iterator n;
	std::vector<BfastTri>::const_iterator t;
	faceNormals.resize(triangles.size());
	vertexNormals.resize(meshPts.size());
	for (n=vertexNormals.begin(); n!=vertexNormals.end(); n++) {
		(*n) = 0.0;
	}
	for (n=faceNormals.begin(),t=triangles.begin(); n!=faceNormals.end(); n++,t++) {
		(*n) = cross(meshPts[(*t)[1]]-meshPts[(*t)[0]], 
								 meshPts[(*t)[2]]-meshPts[(*t)[0]]);
		vertexNormals[(*t)[0]]+=(*n);
		vertexNormals[(*t)[1]]+=(*n);
		vertexNormals[(*t)[2]]+=(*n);
	}

	for (n=vertexNormals.begin(); n!=vertexNormals.end(); n++)
		normalize(*n);
}

void bvtRead(char *fname, std::vector<BfastVector3> &meshPts, std::vector<BfastTri> &triangles) {
	int version;
	ifstream fin (fname, ios::in | ios::binary);
	fin.read((char*)&version, sizeof(int));
	if (version == 1) {
		bfast::read(fin, meshPts);
		bfast::read(fin, triangles);
	} else {
		cerr << "Incompatible bvt file: "<<fname<<endl;
		exit(-1);
	}
	fin.close();
}

void gtDump(char *fname, const std::vector<BfastVector3> &meshPts, const std::vector<BfastTri> &triangles) {
	ofstream out;
	std::vector<BfastVector3>::const_iterator p,n;
	std::vector<BfastTri>::const_iterator t;
	std::vector<BfastVector3> vertexNormals;
	std::vector<BfastVector3> faceNormals;
	computeNormals(meshPts,triangles,vertexNormals,faceNormals);
	out.open(fname);
	out<<"vertices: "<<meshPts.size()<<std::endl;
	out<<"faces: "<<triangles.size()<<std::endl;

	for (p=meshPts.begin(),n=vertexNormals.begin(); p!=meshPts.end(); p++,n++) 
		out<<"v "<<(*p)[0]<<" "<<(*p)[1]<<" "<<(*p)[2]<<" "<<(*n)[0]<<" "<<(*n)[1]<<" "<<(*n)[2]<<std::endl;
	
	for (t=triangles.begin(); t!=triangles.end(); t++) 
		out<<"f "<<(*t)[0]+1<<" "<<(*t)[1]+1<<" "<<(*t)[2]+1<<std::endl;

	out.close();
}

void bgtDump(char *fname, const std::vector<BfastVector3> &meshPts, const std::vector<BfastTri> &triangles) {
	ofstream out;
	std::vector<BfastVector3>::const_iterator p,n;
	std::vector<BfastTri>::const_iterator t;
	std::vector<BfastVector3> vertexNormals;
	std::vector<BfastVector3> faceNormals;
	computeNormals(meshPts,triangles,vertexNormals,faceNormals);
	out.open(fname, ios::out | ios::binary);
	int foo = meshPts.size();
	out.write((char*)&foo, sizeof(int));
	foo = triangles.size();
	out.write((char*)&foo, sizeof(int));
	double bar;
	for (p=meshPts.begin(),n=vertexNormals.begin(); p!=meshPts.end(); p++,n++) {
		bar = (*p)[0];
		out.write((char*)&bar, sizeof(double));
		bar = (*p)[1];
		out.write((char*)&bar, sizeof(double));
		bar = (*p)[2];
		out.write((char*)&bar, sizeof(double));
		bar = (*n)[0];
		out.write((char*)&bar, sizeof(double));
		bar = (*n)[1];
		out.write((char*)&bar, sizeof(double));
		bar = (*n)[2];
		out.write((char*)&bar, sizeof(double));
	}
	for (t=triangles.begin(); t!=triangles.end(); t++) {
		foo = (*t)[0];
		out.write((char*)&foo, sizeof(int));
		foo = (*t)[1];
		out.write((char*)&foo, sizeof(int));
		foo = (*t)[2];
		out.write((char*)&foo, sizeof(int));
	}
	out.close();
}

void gtRead(char *fname, std::vector<BfastVector3> &meshPts, std::vector<BfastTri> &triangles) {
	int i,j;
	ifstream in;
	std::vector<BfastVector3>::iterator p,n;
	std::vector<BfastTri>::iterator t;
	std::vector<BfastVector3> vertexNormals;
	in.open(fname);
	eatStr("vertices:", in);
	in>>i;
	meshPts.resize(i);
	vertexNormals.resize(i);
	eatStr("faces:",in);
	in>>j;
	triangles.resize(j);
	for (p=meshPts.begin(),n=vertexNormals.begin(); p!=meshPts.end(); p++,n++) {
		eatChar('v', in);
		in>>(*p)[0]>>(*p)[1]>>(*p)[2]>>(*n)[0]>>(*n)[1]>>(*n)[2];
	}
	for (t=triangles.begin(); t!=triangles.end(); t++) {
		eatChar('f', in);
		in>>(*t)[0]>>(*t)[1]>>(*t)[2];
		(*t)[0]--;
		(*t)[1]--;
		(*t)[2]--;
	}
	in.close();
}

#if 0
void readObjFile(char *fname, std::vector<BfastVector3> &pts, std::vector<BfastTri> &triangles) {
  ifstream in(fname, ios::in);
  char c[10];
  int vt = 0;
  int vn = 0;
  BfastVector3 lc(BFAST_REAL_MAX), uc(-BFAST_REAL_MAX);
	BfastVector3 x;
	BfastTri t;

  while ( (in >> c) && (strcmp(c,"v") != 0) );
  
  if ( strcmp(c,"v") == 0 ) {
    while ( strcmp(c,"v") == 0 ) {
      //pts.resize(i+1);
      in >> x[0];
      in >> x[1];
      in >> x[2];
      pts.push_back(x)
      if (x[0] < lc[0]) lc[0] = x[0];
      if (x[1] < lc[1]) lc[1] = x[1];
      if (x[2] < lc[2]) lc[2] = x[2];
      if (x[0] > uc[0]) uc[0] = x[0];
      if (x[1] > uc[1]) uc[1] = x[1];
      if (x[2] > uc[2]) uc[2] = x[2];
      
      in >> c;
    }
  }
  
  while ( (strcmp(c,"f") != 0) && (strcmp(c,"vt") != 0) && (strcmp(c,"vn") != 0) && ( in >> c) );
  
  if ( strcmp(c,"vt") == 0 ) {
    vt = 1;
    do {
      BfastReal temp;
      in >> temp;
      in >> temp;
      in >> c;
    } while ( strcmp(c,"vt") == 0 );
  }
  
  while ( (strcmp(c,"f") != 0) && (strcmp(c,"vn") != 0) && ( in >> c) );
  
  if ( strcmp(c,"vn") == 0 ) {
    vn = 1;
    while ( strcmp(c,"vn") == 0 ) {
      BfastReal temp;
      in >> temp;
      in >> temp;
      in >> temp;
      in >> c;
    }
  }
  
  while ( (strcmp(c,"f") != 0) && ( in >> c) );
  if ( strcmp(c,"f") == 0 ) {
    do {
      if ( strcmp(c,"f") != 0 )
				break;
      
      int temp;
      char ct;
      
      in >> t[0];
      if ( vt ) { in >> ct; in >> temp; }
      if ( vn ) { in >> ct; in >> temp; }
      
      in >> t[1];
      if ( vt ) { in >> ct; in >> temp; }
      if ( vn ) { in >> ct; in >> temp; }
      
      in >> t[2];
      if ( vt ) { in >> ct; in >> temp; }
      if ( vn ) { in >> ct; in >> temp; }
      t[0] -= 1;
      t[1] -= 1;
      t[2] -= 1;
			triangles.push_back(t);
    } while ( in >> c );
  }
}
#else
bool readObjFile(char *fname, std::vector<BfastVector3> &pts, std::vector<BfastTri> &triangles) {
  char c[500];
  BfastVector3 lc(BFAST_REAL_MAX), uc(-BFAST_REAL_MAX);
  
  int numVertices=0, numFaces=0;
  bool normals = false, texture = false;
  int tint;
  char ch;
  int p, q, r;
  double x, y, z;
	std::vector<BfastVector3>::iterator v;
	std::vector<BfastTri>::iterator t;
  std::ifstream in1(fname, std::ios::in);
  in1.flags(in1.flags() & ~std::ios::skipws);

  while (in1>>ch) {
    if (ch == 'v') {
      in1>>ch;
      if (ch == ' ') numVertices++;
      else if (ch == 'n') normals = true;
      else if (ch == 't') texture = true;
      else std::cerr<<"error \'"<<ch<<"\'"<<std::endl;
    } else if (ch == 'f') numFaces++;
  }
  in1.close();

  pts.resize(numVertices);
  triangles.resize(numFaces);
	v = pts.begin();
	t = triangles.begin();

  std::ifstream in(fname, std::ios::in);
  while (in>>ch) {
    if (ch == '#') {
      in.getline(c,500);
      continue;
    }
    if (ch == 'g') {
      in.getline(c,500);
      continue;
    }
    if (ch == 's') {
      in.getline(c,500);
      continue;
    }
    if (ch == 'm') {
      in.getline(c,500);
      continue;
    }
    if (ch == 'u') {
      in.getline(c,500);
      continue;
    }
    if (ch == 'v') {
      ch = in.peek();
      if (ch != 't' && ch != 'n') {
				in>>x>>y>>z;
				(*v).set(x,y,z);
				if ((*v)[0] < lc[0]) lc[0] = (*v)[0];
				if ((*v)[1] < lc[1]) lc[1] = (*v)[1];
				if ((*v)[2] < lc[2]) lc[2] = (*v)[2];
				if ((*v)[0] > uc[0]) uc[0] = (*v)[0];
				if ((*v)[1] > uc[1]) uc[1] = (*v)[1];
				if ((*v)[2] > uc[2]) uc[2] = (*v)[2];
				v++;
      } else {
				in.getline(c, 500);
      }
      continue;
    }
    if (ch == 'f') {
      if (normals && texture) {
				in>>p>>ch>>tint>>ch>>tint>>q>>ch>>tint>>ch>>tint>>r>>ch>>tint>>ch>>tint;
      } else if (normals) {
				in>>p>>ch>>ch>>tint>>q>>ch>>ch>>tint>>r>>ch>>ch>>tint;
      } else if (texture) {
				in>>p>>ch>>tint>>q>>ch>>tint>>r>>ch>>tint;
      } else {
				in>>p>>q>>r;
      }
      (*t)[0] = p-1;
      (*t)[1] = q-1;
      (*t)[2] = r-1;
      t++;
      continue;
    }
  }
  in.close();
  return true;
}
#endif


//-------------------------------------------------------------------
//-------------------------------------------------------------------
//
// RCS Revision History
//
// $Log: bfastUtil.cpp,v $
// Revision 1.11  2007/07/31 20:17:38  adamwb
// removed warnings
//
// Revision 1.10  2007/07/26 00:19:51  adamwb
// major updates to stl and faster hash_map usage
//
// Revision 1.9  2007/06/16 01:05:26  adamwb
// *** empty log message ***
//
// Revision 1.8  2007/01/22 19:41:58  cwojtan
// fixed some collision bugs and some problems with loading obstacles in windows
//
// Revision 1.7  2007/01/22 14:04:17  adamwb
// collision object updates
//
// Revision 1.6  2006/12/23 03:09:29  adamwb
// ZZZZ
// 
//
// Revision 1.5  2006/06/16 21:07:13  adamb
// new version
//
// Revision 1.4  2006/01/15 01:43:51  adamb
// added rebuildTree
//
// Revision 1.3  2005/12/29 01:59:38  adamb
// Added void * to all passed in functions
//
// Revision 1.2  2005/11/08 19:45:05  adamb
// forced checkin
//
// Revision 1.1.1.1  2005/09/06 22:45:12  adamb
// imported sources
//
// Revision 1.1  2005/08/10 01:33:25  adamb
// forced
//
// Revision 1.1.1.1  2005/05/25 05:40:59  adamb
// Initial Revision
//
//
//-------------------------------------------------------------------
//-------------------------------------------------------------------

